package cz.jzitnik.game.handlers.place;

import cz.jzitnik.game.annotations.BlockDropPercentage;
import cz.jzitnik.game.annotations.BlockRegistry;
import cz.jzitnik.game.annotations.PlaceOnSolid;
import cz.jzitnik.game.annotations.ResetDataOnMine;
import cz.jzitnik.game.entities.items.ItemBlockSupplier;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.reflections.Reflections;

import java.util.Arrays;
import java.util.Set;

import static org.junit.jupiter.api.Assertions.*;

class PlaceHandlerTest {
    @Test
    @DisplayName("All classes annotated with @PlaceHandler must implement CustomPlaceHandler")
    void allClassesPickupImplementHandler() {
        Reflections reflections = new Reflections("cz.jzitnik.game.handlers.pickup.handlers");
        Set<Class<?>> handlerClasses = reflections.getTypesAnnotatedWith(cz.jzitnik.game.annotations.PlaceHandler.class);

        for (Class<?> clazz : handlerClasses) {
            boolean implementsClass = CustomPlaceHandler.class.isAssignableFrom(clazz);
            assertTrue(implementsClass, "Class " + clazz.getName() + " annotated with @PlaceHandler does not implement CustomPlaceHandler.");
        }
    }


    @Test
    @DisplayName("All blocks in annotation @PlaceHandler must exist")
    void allBlocksMustExist() {
        Reflections reflections = new Reflections("cz.jzitnik.game.handlers.place.handlers");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(cz.jzitnik.game.annotations.PlaceHandler.class);

        for (Class<?> clazz : handlerClasses) {
            cz.jzitnik.game.annotations.PlaceHandler annotation = clazz
                    .getAnnotation(cz.jzitnik.game.annotations.PlaceHandler.class);
            var id = annotation.value();
            assertDoesNotThrow(() -> ItemBlockSupplier.getBlock(id), "Block " + id + " does not exist but is in annotation @PlaceHandler in class " + clazz.getName());
        }
    }

    @Test
    @DisplayName("All blocks annotated with @PlaceOnSolid must be also annotated with @BlockRegistry")
    void placeOnSolidTest() {
        Reflections reflections = new Reflections("cz.jzitnik.game.entities.items.registry.blocks");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(PlaceOnSolid.class);

        for (Class<?> clazz : handlerClasses) {
            boolean hasAnnotation = clazz.isAnnotationPresent(BlockRegistry.class);
            assertTrue(hasAnnotation, "Class " + clazz.getName() + " is annotated with @PlaceOnSolid but does not have annotation @BlockRegistry");
        }
    }

    @Test
    @DisplayName("All blocks annotated with @ResetDataOnMine must be also annotated with @BlockRegistry")
    void resetDataOnMineTest() {
        Reflections reflections = new Reflections("cz.jzitnik.game.entities.items.registry.blocks");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(ResetDataOnMine.class);

        for (Class<?> clazz : handlerClasses) {
            boolean hasAnnotation = clazz.isAnnotationPresent(BlockRegistry.class);
            assertTrue(hasAnnotation, "Class " + clazz.getName() + " is annotated with @ResetDataOnMine but does not have annotation @BlockRegistry");
        }
    }

    @Test
    @DisplayName("All blocks annotated with @BlockDropPercentage must be also annotated with @BlockRegistry")
    void blockDropPercentageTest() {
        Reflections reflections = new Reflections("cz.jzitnik.game.entities.items.registry.blocks");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(BlockDropPercentage.class);

        for (Class<?> clazz : handlerClasses) {
            boolean hasAnnotation = clazz.isAnnotationPresent(BlockRegistry.class);
            assertTrue(hasAnnotation, "Class " + clazz.getName() + " is annotated with @BlockDropPercentage but does not have annotation @BlockRegistry");
        }
    }

    @Test
    @DisplayName("All blocks annotated with @ResetDataOnMine must have data class with default constructor")
    void dataOnMineHasDataTest() {
        Reflections reflections = new Reflections("cz.jzitnik.game.entities.items.registry.blocks");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(ResetDataOnMine.class);

        for (Class<?> clazz : handlerClasses) {
            var annotation = clazz.getAnnotation(BlockRegistry.class);
            var block = ItemBlockSupplier.getBlock(annotation.value());
            assertNotNull(block.getData(), "Block " + clazz.getName() + " annotated with @ResetDataOnMine does not have default Data set.");
        }
    }

    @Test
    @DisplayName("All blocks annotated with @ResetDataOnMine must have data class with default constructor")
    void dataOnMineDefaultConstructorTest() {
        Reflections reflections = new Reflections("cz.jzitnik.game.entities.items.registry.blocks");
        Set<Class<?>> handlerClasses = reflections
                .getTypesAnnotatedWith(ResetDataOnMine.class);

        for (Class<?> clazz : handlerClasses) {
            var annotation = clazz.getAnnotation(BlockRegistry.class);
            var block = ItemBlockSupplier.getBlock(annotation.value());
            var dataClass = block.getData().getClass();
            var hasDefaultConstructor = Arrays.stream(dataClass.getDeclaredConstructors()).anyMatch(constructor -> constructor.getParameterCount() == 0);
            assertTrue(hasDefaultConstructor, "Block " + clazz.getName() + " annotated with @ResetDataOnMine does not have Data that has default constructor.");
        }
    }
}